from surya.detection import batch_text_detection
from surya.layout import batch_layout_detection
from surya.postprocessing.util import rescale_bbox
from surya.schema import Bbox


def merge_boxes(box1, box2):
    return [min(box1[0], box2[0]), min(box1[1], box2[1]), max(box1[2], box2[2]), max(box1[3], box2[3])]


def merge_tables(page_table_boxes):
    # Merge tables that are next to each other
    expansion_factor = 1.02
    shrink_factor = .98
    ignore_boxes = set()
    for i in range(len(page_table_boxes)):
        if i in ignore_boxes:
            continue
        for j in range(i + 1, len(page_table_boxes)):
            if j in ignore_boxes:
                continue
            expanded_box1 = [page_table_boxes[i][0] * shrink_factor, page_table_boxes[i][1],
                             page_table_boxes[i][2] * expansion_factor, page_table_boxes[i][3]]
            expanded_box2 = [page_table_boxes[j][0] * shrink_factor, page_table_boxes[j][1],
                             page_table_boxes[j][2] * expansion_factor, page_table_boxes[j][3]]
            if Bbox(bbox=expanded_box1).intersection_pct(Bbox(bbox=expanded_box2)) > 0:
                page_table_boxes[i] = merge_boxes(page_table_boxes[i], page_table_boxes[j])
                ignore_boxes.add(j)

    return [b for i, b in enumerate(page_table_boxes) if i not in ignore_boxes]


def detect_tables(images, highres_images, models):
    det_model, det_processor, layout_model, layout_processor = models
    line_predictions = batch_text_detection(images, det_model, det_processor)
    layout_predictions = batch_layout_detection(images, layout_model, layout_processor, line_predictions)

    table_imgs = []
    table_counts = []
    table_bboxes = []

    for layout_pred, img, highres_img in zip(layout_predictions, images, highres_images):
        # The bbox for the entire table
        bbox = [l.bbox for l in layout_pred.bboxes if l.label == "Table"]

        if len(bbox) == 0:
            table_counts.append(0)
            continue

        page_table_imgs = []
        highres_bbox = []

        # Merge tables that are next to each other
        bbox = merge_tables(bbox)

        # Number of tables per page
        table_counts.append(len(bbox))

        for bb in bbox:
            highres_bb = rescale_bbox(bb, img.size, highres_img.size)
            page_table_imgs.append(highres_img.crop(highres_bb))
            highres_bbox.append(highres_bb)

        table_imgs.extend(page_table_imgs)
        table_bboxes.extend(highres_bbox)

    return table_imgs, table_bboxes, table_counts


import re
from typing import List

from tabled.schema import SpanTableCell


def sort_within_cell(cells, tolerance=5):
    vertical_groups = {}
    for i, cell in enumerate(cells):
        group_key = round((cell.bbox[1] + cell.bbox[3]) / 2 / tolerance)
        if group_key not in vertical_groups:
            vertical_groups[group_key] = []
        vertical_groups[group_key].append((i, cell.bbox[0]))

    # Sort each group horizontally and flatten the groups into a single list
    sorted_cell_idxs = []
    for _, group in sorted(vertical_groups.items()):
        sorted_group = sorted(group, key=lambda x: x[1])
        sorted_cell_idxs.extend([idx for idx, _ in sorted_group])

    cell_order = [sorted_cell_idxs.index(i) for i in range(len(sorted_cell_idxs))]
    return cell_order


def sort_cells(cells: List[SpanTableCell]):
    cell_order = sort_within_cell(cells)
    for i, cell in enumerate(cells):
        cell.order = cell_order[i]
    cells.sort(key=lambda x: (x.row_ids[0], x.col_ids[0], x.order))
    return cells


def replace_dots(text):
    dot_pattern = re.compile(r'(\s*\.\s*){4,}')
    dot_multiline_pattern = re.compile(r'.*(\s*\.\s*){4,}.*', re.DOTALL)

    if dot_multiline_pattern.match(text):
        text = dot_pattern.sub(' ', text)
    return text


def replace_newlines(text):
    # Replace all newlines
    newline_pattern = re.compile(r'[\r\n]+')
    return newline_pattern.sub(' ', text).strip()



from typing import List

from tabulate import tabulate

from tabled.formats.common import sort_cells, replace_dots, replace_newlines
from tabled.schema import SpanTableCell
import csv
import io


def replace_all(text):
    return replace_newlines(replace_dots(text))


def csv_format(cells: List[SpanTableCell]):
    cells = sort_cells(cells)
    unique_rows = set([cell.row_ids[0] for cell in cells])
    unique_cols = set([cell.col_ids[0] for cell in cells])
    buff = io.StringIO()
    writer = csv.writer(buff)
    for row in unique_rows:
        text_row = []
        for col in unique_cols:
            cell = " ".join([cell.text for cell in cells if cell.row_ids[0] == row and cell.col_ids[0] == col])
            cell = replace_all(cell)
            text_row.append(cell)
        writer.writerow(text_row)

    csv_str = buff.getvalue()
    return csv_str



import numpy as np
from sklearn.cluster import DBSCAN


def cluster_coords(coords, row_count):
    if len(coords) == 0:
        return []
    coords = np.array(sorted(set(coords))).reshape(-1, 1)

    clustering = DBSCAN(eps=.01, min_samples=max(2, row_count // 4)).fit(coords)
    clusters = clustering.labels_

    separators = []
    for label in set(clusters):
        clustered_points = coords[clusters == label]
        separators.append(np.mean(clustered_points))

    separators = sorted(separators)
    return separators


def find_column_separators(rows, page_size, round_factor=.002, min_count=1):
    left_edges = []
    right_edges = []
    centers = []

    boxes = [c.bbox for r in rows for c in r]

    for cell in boxes:
        ncell = [cell[0] / page_size[0], cell[1] / page_size[1], cell[2] / page_size[0], cell[3] / page_size[1]]
        left_edges.append(ncell[0] / round_factor * round_factor)
        right_edges.append(ncell[2] / round_factor * round_factor)
        centers.append((ncell[0] + ncell[2]) / 2 * round_factor / round_factor)

    left_edges = [l for l in left_edges if left_edges.count(l) > min_count]
    right_edges = [r for r in right_edges if right_edges.count(r) > min_count]
    centers = [c for c in centers if centers.count(c) > min_count]

    sorted_left = cluster_coords(left_edges, len(rows))
    sorted_right = cluster_coords(right_edges, len(rows))
    sorted_center = cluster_coords(centers, len(rows))

    # Find list with minimum length
    separators = max([sorted_left, sorted_right, sorted_center], key=len)
    separators.append(1)
    separators.insert(0, 0)
    return separators


def assign_cells_to_columns(rows, page_size, round_factor=.002, tolerance=.01):
    separators = find_column_separators(rows, page_size, round_factor=round_factor)
    additional_column_index = 0
    row_dicts = []

    for row in rows:
        new_row = {}
        last_col_index = -1
        for cell in row:
            left_edge = cell.bbox[0] / page_size[0]
            column_index = -1
            for i, separator in enumerate(separators):
                if left_edge - tolerance < separator and last_col_index < i:
                    column_index = i
                    break
            if column_index == -1:
                column_index = len(separators) + additional_column_index
                additional_column_index += 1
            new_row[column_index] = cell
            last_col_index = column_index
        additional_column_index = 0
        row_dicts.append(new_row)

    cells = []
    for row_idx, row in enumerate(row_dicts):
        column = 0
        for col_idx in sorted(row.keys()):
            cell = row[col_idx]
            cell.row_ids = [row_idx]
            cell.col_ids = [column]
            cells.append(cell)
            column += 1

    return cells


import os

from surya.input.load import load_from_folder, load_from_file
from surya.settings import settings as surya_settings


def load_pdfs_images(input_path, max_pages=None, start_page=None):
    if os.path.isdir(input_path):
        images, _, _ = load_from_folder(input_path, max_pages, start_page=start_page)
        highres_images, names, text_lines = load_from_folder(input_path, max_pages, dpi=surya_settings.IMAGE_DPI_HIGHRES,
                                                             load_text_lines=True, start_page=start_page)
    else:
        images, _, _ = load_from_file(input_path, max_pages, start_page=start_page)
        highres_images, names, text_lines = load_from_file(input_path, max_pages, dpi=surya_settings.IMAGE_DPI_HIGHRES,
                                                           load_text_lines=True, start_page=start_page)

    return images, highres_images, names, text_lines


from typing import Optional

from dotenv import find_dotenv
from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    # General
    IN_STREAMLIT: bool = False
    TORCH_DEVICE: Optional[str] = None

    class Config:
        env_file = find_dotenv("local.env")
        extra = "ignore"


settings = Settings()


import os

from tabled.assignment import assign_rows_columns
from tabled.fileinput import load_pdfs_images
from tabled.formats.markdown import markdown_format
from tabled.inference.detection import detect_tables
from tabled.inference.recognition import get_cells, recognize_tables

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
os.environ["IN_STREAMLIT"] = "true"
import pypdfium2

import io
import tempfile
from PIL import Image

import streamlit as st
from tabled.inference.models import load_detection_models, load_recognition_models


@st.cache_resource()
def load_models():
    return load_detection_models(), load_recognition_models()


def run_table_rec(image, highres_image, text_line, models, skip_detection=False, detect_boxes=False):
    if not skip_detection:
        table_imgs, table_bboxes, _ = detect_tables([image], [highres_image], models[0])
    else:
        table_imgs = [highres_image]
        table_bboxes = [[0, 0, highres_image.size[0], highres_image.size[1]]]

    table_text_lines = [text_line] * len(table_imgs)
    highres_image_sizes = [highres_image.size] * len(table_imgs)
    cells, needs_ocr = get_cells(table_imgs, table_bboxes, highres_image_sizes, table_text_lines, models[0][:2], detect_boxes=detect_boxes)

    table_rec = recognize_tables(table_imgs, cells, needs_ocr, models[1])
    cells = [assign_rows_columns(tr, im_size) for tr, im_size in zip(table_rec, highres_image_sizes)]

    out_data = []
    for idx, (cell, pred, table_img) in enumerate(zip(cells, table_rec, table_imgs)):
        md = markdown_format(cell)
        out_data.append((md, table_img))
    return out_data


def open_pdf(pdf_file):
    stream = io.BytesIO(pdf_file.getvalue())
    return pypdfium2.PdfDocument(stream)


@st.cache_data()
def get_page_image(pdf_file, page_num, dpi=96):
    doc = open_pdf(pdf_file)
    renderer = doc.render(
        pypdfium2.PdfBitmap.to_pil,
        page_indices=[page_num - 1],
        scale=dpi / 72,
    )
    png = list(renderer)[0]
    png_image = png.convert("RGB")
    return png_image


@st.cache_data()
def page_count(pdf_file):
    doc = open_pdf(pdf_file)
    return len(doc)


st.set_page_config(layout="wide")

models = load_models()


st.markdown("""
# Tabled Demo

This app will let you try tabled, a table detection and recognition model.  It will detect and recognize the tables.

Find the project [here](https://github.com/VikParuchuri/tabled).
""")

in_file = st.sidebar.file_uploader("PDF file or image:", type=["pdf", "png", "jpg", "jpeg", "gif", "webp"])
skip_detection = st.sidebar.checkbox("Skip table detection", help="Use this if tables are already cropped (the whole PDF page or image is a table)", value=False)
detect_boxes = st.sidebar.checkbox("Detect cell boxes", help="Detect table cell boxes vs extract from PDF.  Will also run OCR.", value=False)

if in_file is None:
    st.stop()

filetype = in_file.type
col = st.columns(1)[0]
container = col.container()

if "pdf" in filetype:
    page_count = page_count(in_file)
    page_number = st.sidebar.number_input(f"Page number out of {page_count}:", min_value=1, value=1,
                                          max_value=page_count)

    pil_image = get_page_image(in_file, page_number, 96)
else:
    pil_image = Image.open(in_file).convert("RGB")
    pil_image_highres = pil_image
    page_number = 1

with col:
    st.image(pil_image, caption="PDF file (preview)", use_column_width="auto")

    run_marker = st.sidebar.button("Run Tabled")

if not run_marker:
    st.stop()

# Run Tabled
file_ext = in_file.name.rsplit(".")[-1]
with tempfile.NamedTemporaryFile(suffix=file_ext) as temp_input:
    temp_input.write(in_file.getvalue())
    temp_input.seek(0)
    filename = temp_input.name
    images, highres_images, names, text_lines = load_pdfs_images(filename, max_pages=1, start_page=page_number - 1)
    out_data = run_table_rec(images[0], highres_images[0], text_lines[0], models, skip_detection=skip_detection, detect_boxes=detect_boxes)

for idx, (md, table_img) in enumerate(out_data):
    container.markdown(f"## Table {idx}")
    container.image(table_img, caption=f"Table {idx}", use_column_width="auto")
    container.markdown(md)
    container.code(md)
    container.divider()




from typing import List

from tabled.assignment import assign_rows_columns
from tabled.inference.detection import detect_tables
from tabled.inference.recognition import get_cells, recognize_tables
from tabled.schema import ExtractPageResult


def extract_tables(images, highres_images, text_lines, det_models, rec_models, skip_detection=False, detect_boxes=False) -> List[ExtractPageResult]:
    if not skip_detection:
        table_imgs, table_bboxes, table_counts = detect_tables(images, highres_images, det_models)
    else:
        table_imgs = highres_images
        table_bboxes = [[0, 0, img.size[0], img.size[1]] for img in highres_images]
        table_counts = [1] * len(highres_images)

    table_text_lines = []
    highres_image_sizes = []
    for i, tc in enumerate(table_counts):
        table_text_lines.extend([text_lines[i]] * tc)
        highres_image_sizes.extend([highres_images[i].size] * tc)

    cells, needs_ocr = get_cells(table_imgs, table_bboxes, highres_image_sizes, table_text_lines, det_models[:2], detect_boxes=detect_boxes)

    table_rec = recognize_tables(table_imgs, cells, needs_ocr, rec_models)
    cells = [assign_rows_columns(tr, im_size) for tr, im_size in zip(table_rec, highres_image_sizes)]

    results = []
    counter = 0
    for count in table_counts:
        page_start = counter
        page_end = counter + count
        results.append(ExtractPageResult(
            table_imgs=table_imgs[page_start:page_end],
            cells=cells[page_start:page_end],
            rows_cols=table_rec[page_start:page_end]
        ))
        counter += count

    assert len(results) == len(images)
    return results



from typing import List

from tabulate import tabulate

from tabled.formats.common import sort_cells, replace_dots, replace_newlines
from tabled.schema import SpanTableCell


def replace_special_chars(text):
    return text.replace("|", "\|").replace("-", "\-")


def replace_all(text):
    return replace_special_chars(replace_newlines(replace_dots(text)))


def markdown_format(cells: List[SpanTableCell]):
    md_rows = []
    cells = sort_cells(cells)
    unique_rows = set([cell.row_ids[0] for cell in cells])
    unique_cols = set([cell.col_ids[0] for cell in cells])
    for row in unique_rows:
        md_row = []
        for col in unique_cols:
            cell = " ".join([cell.text for cell in cells if cell.row_ids[0] == row and cell.col_ids[0] == col])
            cell = replace_all(cell)
            md_row.append(cell)
        md_rows.append(md_row)

    md = tabulate(md_rows, headers="firstrow", tablefmt="github", disable_numparse=True)
    return md



from typing import List

import numpy as np
from surya.schema import TableResult, Bbox

from tabled.heuristics import heuristic_layout
from tabled.schema import SpanTableCell


def is_rotated(rows, cols):
    # Determine if the table is rotated by looking at row and column width / height ratios
    # Rows should have a >1 ratio, cols <1
    widths = sum([r.width for r in rows])
    heights = sum([c.height for c in rows]) + 1
    r_ratio = widths / heights

    widths = sum([c.width for c in cols])
    heights = sum([r.height for r in cols]) + 1
    c_ratio = widths / heights

    return r_ratio * 2 < c_ratio


def overlapper_idxs(rows, field, thresh=.3):
    overlapper_rows = set()
    for row in rows:
        row_id = getattr(row, field)
        if row_id in overlapper_rows:
            continue

        for row2 in rows:
            row2_id = getattr(row2, field)
            if row2_id == row_id or row2_id in overlapper_rows:
                continue

            if row.intersection_pct(row2) > thresh:
                i_bigger = row.area > row2.area
                overlapper_rows.add(row_id if i_bigger else row2_id)
    return overlapper_rows


def initial_assignment(detection_result: TableResult, thresh=.5) -> List[SpanTableCell]:
    overlapper_rows = overlapper_idxs(detection_result.rows, field="row_id")
    overlapper_cols = overlapper_idxs(detection_result.cols, field="col_id")

    cells = []
    for cell in detection_result.cells:
        max_intersection = 0
        row_pred = None
        for row in detection_result.rows:
            if row.row_id in overlapper_rows:
                continue

            intersection_pct = Bbox(bbox=cell.bbox).intersection_pct(row)
            if intersection_pct > max_intersection and intersection_pct > thresh:
                max_intersection = intersection_pct
                row_pred = row.row_id

        max_intersection = 0
        col_pred = None
        for col in detection_result.cols:
            if col.col_id in overlapper_cols:
                continue

            intersection_pct = Bbox(bbox=cell.bbox).intersection_pct(col)
            if intersection_pct > max_intersection and intersection_pct > thresh:
                max_intersection = intersection_pct
                col_pred = col.col_id

        cells.append(
            SpanTableCell(
                bbox=cell.bbox,
                text=cell.text,
                row_ids=[row_pred],
                col_ids=[col_pred]
            )
        )
    return cells


def assign_overlappers(cells: List[SpanTableCell], detection_result: TableResult, thresh=.5):
    overlapper_rows = overlapper_idxs(detection_result.rows, field="row_id")
    overlapper_cols = overlapper_idxs(detection_result.cols, field="col_id")

    for cell in cells:
        max_intersection = 0
        row_pred = None
        for row in detection_result.rows:
            if row.row_id not in overlapper_rows:
                continue

            intersection_pct = Bbox(bbox=cell.bbox).intersection_pct(row)
            if intersection_pct > max_intersection and intersection_pct > thresh:
                max_intersection = intersection_pct
                row_pred = row.row_id

        max_intersection = 0
        col_pred = None
        for col in detection_result.cols:
            if col.col_id not in overlapper_cols:
                continue

            intersection_pct = Bbox(bbox=cell.bbox).intersection_pct(col)
            if intersection_pct > max_intersection and intersection_pct > thresh:
                max_intersection = intersection_pct
                col_pred = col.col_id

        if cell.row_ids[0] is None:
            cell.row_ids = [row_pred]
        if cell.col_ids[0] is None:
            cell.col_ids = [col_pred]


def assign_unassigned(table_cells: list, detection_result: TableResult):
    rotated = is_rotated(detection_result.rows, detection_result.cols)
    for cell in table_cells:
        if cell.row_ids[0] is None:
            closest_row = None
            min_dist = None
            for row in detection_result.rows:
                if rotated:
                    dist = cell.center_x_distance(row)
                else:
                    dist = cell.center_y_distance(row)

                if min_dist is None or dist < min_dist:
                    closest_row = row.row_id
                    min_dist = dist
            cell.row_ids = [closest_row]

        if cell.col_ids[0] is None:
            closest_col = None
            min_dist = None
            for col in detection_result.cols:
                if rotated:
                    dist = cell.center_y_distance(col)
                else:
                    dist = cell.center_x_distance(col)

                if min_dist is None or dist < min_dist:
                    closest_col = col.col_id
                    min_dist = dist

            cell.col_ids = [closest_col]


def handle_rowcol_spans(table_cells: list, detection_result: TableResult, thresh=.25):
    rotated = is_rotated(detection_result.rows, detection_result.cols)
    for cell in table_cells:
        for c in detection_result.cols:
            col_intersect_pct = cell.intersection_y_pct(c) if rotated else cell.intersection_x_pct(c)
            other_cell_exists = len([tc for tc in table_cells if tc.col_ids[0] == c.col_id and tc.row_ids[0] == cell.row_ids[0]]) > 0
            if col_intersect_pct > thresh and not other_cell_exists:
                cell.col_ids.append(c.col_id)
            else:
                break
        # Assign to first column header appears in
        cell.col_ids = sorted(cell.col_ids)

    for cell in table_cells:
        for r in detection_result.rows:
            row_intersect_pct = cell.intersection_x_pct(r) if rotated else cell.intersection_y_pct(r)
            other_cell_exists = len([tc for tc in table_cells if tc.row_ids[0] == r.row_id and tc.col_ids[0] == cell.col_ids[0]]) > 0
            if row_intersect_pct > thresh and not other_cell_exists:
                cell.row_ids.append(r.row_id)
            else:
                break


def merge_multiline_rows(detection_result: TableResult, table_cells: List[SpanTableCell]):
    def find_row_gap(r1, r2):
        return min([abs(r1.bbox[1] - r2.bbox[3]), abs(r2.bbox[1] - r1.bbox[3])])

    all_cols = set([tc.col_ids[0] for tc in table_cells])
    if len(all_cols) == 0:
        return

    merged_pairs = []
    row_gaps = [
        find_row_gap(r, r2)
        for r, r2 in zip(detection_result.rows, detection_result.rows[1:])
    ]
    if len(row_gaps) == 0:
        return

    gap_thresh = np.median(row_gaps)

    for idx, row in enumerate(detection_result.rows[1:]):
        prev_row = detection_result.rows[idx - 1]
        gap = find_row_gap(prev_row, row)

        # Ensure the gap between r2 and r1 is small
        if gap > gap_thresh:
            continue

        r1_cells = [tc for tc in table_cells if tc.row_ids[0] == prev_row.row_id]
        r2_cells = [tc for tc in table_cells if tc.row_ids[0] == row.row_id]
        r1_cols = set([tc.col_ids[0] for tc in r1_cells])
        r2_cols = set([tc.col_ids[0] for tc in r2_cells])

        # Ensure all columns in r2 are in r1
        if len(r2_cols - r1_cols) > 0:
            continue

        # Ensure r2 has mostly blank cells
        if len(r2_cols) / len(all_cols) > .5:
            continue

        merged_pairs.append((idx - 1, idx))

    to_remove = set()
    for r1_idx, r2_idx in merged_pairs:
        detection_result.rows[r1_idx].bbox = [
            min(detection_result.rows[r1_idx].bbox[0], detection_result.rows[r2_idx].bbox[0]),
            min(detection_result.rows[r1_idx].bbox[1], detection_result.rows[r2_idx].bbox[1]),
            max(detection_result.rows[r1_idx].bbox[2], detection_result.rows[r2_idx].bbox[2]),
            max(detection_result.rows[r1_idx].bbox[3], detection_result.rows[r2_idx].bbox[3])
        ]
        to_remove.add(r2_idx)

    new_rows = []
    row_counter = 0
    for idx, row in enumerate(detection_result.rows):
        if idx not in to_remove:
            row.row_id = row_counter
            new_rows.append(row)
            row_counter += 1
    detection_result.rows = new_rows


def assign_rows_columns(detection_result: TableResult, image_size: list, heuristic_thresh=.6) -> List[SpanTableCell]:
    table_cells = initial_assignment(detection_result)
    merge_multiline_rows(detection_result, table_cells)
    table_cells = initial_assignment(detection_result)
    assign_overlappers(table_cells, detection_result)
    total_unassigned = len([tc for tc in table_cells if tc.row_ids[0] is None or tc.col_ids[0] is None])
    unassigned_frac = total_unassigned / max(len(table_cells), 1)

    if unassigned_frac > heuristic_thresh:
        table_cells = heuristic_layout(table_cells, image_size)
        return table_cells

    assign_unassigned(table_cells, detection_result)
    handle_rowcol_spans(table_cells, detection_result)
    return table_cells



from tabled.formats.csv import csv_format
from tabled.formats.html import html_format
from tabled.formats.markdown import markdown_format


def formatter(format, page_cells):
    if format == "csv":
        return csv_format(page_cells), "csv"
    elif format == "markdown":
        return markdown_format(page_cells), "md"
    elif format == "html":
        return html_format(page_cells), "html"
    else:
        raise ValueError(f"Invalid format: {format}")


from typing import List

from tabled.heuristics.cells import assign_cells_to_columns
from tabled.schema import SpanTableCell


def heuristic_layout(table_cells: List[SpanTableCell], page_size, row_tol=.01) -> List[SpanTableCell]:
    table_rows = []
    table_row = []
    y_top = None
    y_bottom = None
    for cell in table_cells:
        normed_y_start = cell.bbox[1] / page_size[1]
        normed_y_end = cell.bbox[3] / page_size[1]

        if y_top is None:
            y_top = normed_y_start
        if y_bottom is None:
            y_bottom = normed_y_end

        y_dist = min(abs(normed_y_start - y_bottom), abs(normed_y_end - y_bottom))
        if y_dist < row_tol:
            table_row.append(cell)
        else:
            # New row
            if len(table_row) > 0:
                table_rows.append(table_row)
            table_row = [cell]
            y_top = normed_y_start
            y_bottom = normed_y_end
    if len(table_row) > 0:
        table_rows.append(table_row)

    return assign_cells_to_columns(table_rows, page_size)


from typing import List, Optional, Any

from pydantic import BaseModel, model_validator
from surya.schema import Bbox, TableResult


def str_join(list, char=","):
    return char.join([str(x) for x in list])


class SpanTableCell(Bbox):
    text: str
    row_ids: List[Optional[int]]
    col_ids: List[Optional[int]]
    order: Optional[int] = None

    def intersection_x_pct(self, other):
        if self.width == 0:
            return 0

        x_overlap = max(0, min(self.bbox[2], other.bbox[2]) - max(self.bbox[0], other.bbox[0]))
        return x_overlap / self.width


    def intersection_y_pct(self, other):
        if self.height == 0:
            return 0

        y_overlap = max(0, min(self.bbox[3], other.bbox[3]) - max(self.bbox[1], other.bbox[1]))
        return y_overlap / self.height

    @property
    def label(self):
        return f"{str_join(self.row_ids)}-{str_join(self.col_ids)}"

    def center_x_distance(self, other):
        return abs(self.center[0] - other.center[0])

    def center_y_distance(self, other):
        return abs(self.center[1] - other.center[1])


class ExtractPageResult(BaseModel):
    cells: List[List[SpanTableCell]]
    rows_cols: List[TableResult]
    table_imgs: List[Any]

    @model_validator(mode="after")
    def check_cells(self):
        assert len(self.cells) == len(self.table_imgs), "Cells and table images must be the same length"
        assert len(self.cells) == len(self.rows_cols), "Cells and rows/cols must be the same length"
        return self

    @property
    def total(self):
        return len(self.cells)



from collections import defaultdict

import pypdfium2
import copy
import os

import click
from surya.postprocessing.heatmap import draw_bboxes_on_image

from tabled.extract import extract_tables
from tabled.formats import formatter
from tabled.fileinput import load_pdfs_images
from tabled.inference.models import load_detection_models, load_recognition_models


@click.command(help="Extract tables from PDFs")
@click.argument("in_path", type=click.Path(exists=True))
@click.argument("out_folder", type=click.Path())
@click.option("--save_json", is_flag=True, help="Save row/column/cell information in json format")
@click.option("--save_debug_images", is_flag=True, help="Save images for debugging")
@click.option("--skip_detection", is_flag=True, help="Skip table detection")
@click.option("--detect_cell_boxes", is_flag=True, help="Detect table cell boxes vs extract from PDF.  Will also run OCR.")
@click.option("--format", type=click.Choice(["markdown", "csv", "html"]), default="markdown")
def main(in_path, out_folder, save_json, save_debug_images, skip_detection, detect_cell_boxes, format):
    os.makedirs(out_folder, exist_ok=True)
    images, highres_images, names, text_lines = load_pdfs_images(in_path)
    pnums = []
    prev_name = None
    for i, name in enumerate(names):
        if prev_name is None or prev_name != name:
            pnums.append(0)
        else:
            pnums.append(pnums[-1] + 1)

        prev_name = name

    det_models = load_detection_models()
    rec_models = load_recognition_models()

    page_results = extract_tables(images, highres_images, text_lines, det_models, rec_models, skip_detection=skip_detection, detect_boxes=detect_cell_boxes)

    out_json = defaultdict(list)
    for name, pnum, result in zip(names, pnums, page_results):
        for i in range(result.total):
            page_cells = result.cells[i]
            page_rc = result.rows_cols[i]
            img = result.table_imgs[i]

            base_path = os.path.join(out_folder, name)
            os.makedirs(base_path, exist_ok=True)

            formatted_result, ext = formatter(format, page_cells)
            base_name = f"page{pnum}_table{i}"
            with open(os.path.join(base_path, f"{base_name}.{ext}"), "w") as f:
                f.write(formatted_result)

            img.save(os.path.join(base_path, f"{base_name}.png"))

            if save_json:
                result = {
                    "cells": [c.model_dump() for c in page_cells],
                    "rows": [r.model_dump() for r in page_rc.rows],
                    "cols": [c.model_dump() for c in page_rc.cols]
                }
                out_json[name].append(result)

            if save_debug_images:
                boxes = [l.bbox for l in page_cells]
                labels = [l.label for l in page_cells]
                bbox_image = draw_bboxes_on_image(boxes, copy.deepcopy(img), labels=labels, label_font_size=20)
                bbox_image.save(os.path.join(base_path, f"{base_name}_cells.png"))

                rows = [l.bbox for l in page_rc.rows]
                cols = [l.bbox for l in page_rc.cols]
                row_labels = [f"Row {l.row_id}" for l in page_rc.rows]
                col_labels = [f"Col {l.col_id}" for l in page_rc.cols]

                rc_image = copy.deepcopy(img)
                rc_image = draw_bboxes_on_image(rows, rc_image, labels=row_labels, label_font_size=20, color="blue")
                rc_image = draw_bboxes_on_image(cols, rc_image, labels=col_labels, label_font_size=20, color="red")
                rc_image.save(os.path.join(base_path, f"{base_name}_rc.png"))


if __name__ == "__main__":
    main()



from surya.model.detection.model import load_model as load_det_model, load_processor as load_det_processor
from surya.model.recognition.model import load_model as load_rec_model
from surya.model.recognition.processor import load_processor as load_rec_processor
from surya.model.table_rec.model import load_model as load_table_rec_model
from surya.model.table_rec.processor import load_processor as load_table_rec_processor
from surya.settings import settings as surya_settings


def load_detection_models():
    det_model = load_det_model()
    det_processor = load_det_processor()
    layout_model = load_det_model(checkpoint=surya_settings.LAYOUT_MODEL_CHECKPOINT)
    layout_processor = load_det_processor(checkpoint=surya_settings.LAYOUT_MODEL_CHECKPOINT)
    return det_model, det_processor, layout_model, layout_processor


def load_recognition_models():
    table_rec_model = load_table_rec_model()
    table_rec_processor = load_table_rec_processor()
    rec_model = load_rec_model()
    rec_processor = load_rec_processor()
    return table_rec_model, table_rec_processor, rec_model, rec_processor


from typing import List

from surya.detection import batch_text_detection
from surya.input.pdflines import get_table_blocks
from surya.ocr import run_recognition
from surya.schema import TableResult
from surya.tables import batch_table_recognition


def get_cells(table_imgs, table_bboxes, image_sizes, text_lines, models, detect_boxes=False):
    det_model, det_processor = models
    table_cells = []
    needs_ocr = []

    to_inference_idxs = []
    for idx, (highres_bbox, text_line, image_size) in enumerate(zip(table_bboxes, text_lines, image_sizes)):
        # The text cells inside each table
        table_blocks = get_table_blocks([highres_bbox], text_line, image_size)[0] if text_line is not None else None

        if text_line is None or detect_boxes or len(table_blocks) == 0:
            to_inference_idxs.append(idx)
            table_cells.append(None)
            needs_ocr.append(True)
        else:
            table_cells.append(table_blocks)
            needs_ocr.append(False)

    # Inference tables that need it
    if len(to_inference_idxs) > 0:
        det_results = batch_text_detection([table_imgs[i] for i in to_inference_idxs], det_model, det_processor)
        for idx, det_result in zip(to_inference_idxs, det_results):
            cell_bboxes = [{"bbox": tb.bbox, "text": None} for tb in det_result.bboxes]
            table_cells[idx] = cell_bboxes

    return table_cells, needs_ocr


def recognize_tables(table_imgs, table_cells, needs_ocr: List[bool], models) -> List[TableResult]:
    table_rec_model, table_rec_processor, ocr_model, ocr_processor = models

    if sum(needs_ocr) > 0:
        needs_ocr_idx = [idx for idx, needs in enumerate(needs_ocr) if needs]
        ocr_images = [img for img, needs in zip(table_imgs, needs_ocr) if needs]
        ocr_cells = [[c["bbox"] for c in cells] for cells, needs in zip(table_cells, needs_ocr) if needs]
        ocr_langs = [None] * len(ocr_images)

        ocr_predictions = run_recognition(ocr_images, ocr_langs, ocr_model, ocr_processor, bboxes=ocr_cells)

        # Assign text to correct spot
        for orig_idx, ocr_pred in zip(needs_ocr_idx, ocr_predictions):
            for ocr_line, cell in zip(ocr_pred.text_lines, table_cells[orig_idx]):
                cell["text"] = ocr_line.text

    table_preds = batch_table_recognition(table_imgs, table_cells, table_rec_model, table_rec_processor)
    return table_preds


